raise DeprecationWarning(
    "ray.util.lightgbm has been removed as of Ray 2.0. Instead, use the `lightgbm-ray` "
    "library directly or the `LightGBMTrainer` in Ray AIR ("
    "https://docs.ray.io/en/master/ray-air/getting-started.html)"
)
