# TODO(sven):
#  Add a simple script that takes n csv input files and generates plot(s)
#  from these with: x-axis=ts OR wall-time; y-axis=any metric(s) (up to 2).
#  ability to merge any m csv files (e.g. tf vs torch; or n seeds) together
#  in one plot.
