# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opencensus/proto/resource/v1/resource.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+opencensus/proto/resource/v1/resource.proto\x12\x1copencensus.proto.resource.v1\"\xa5\x01\n\x08Resource\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12J\n\x06labels\x18\x02 \x03(\x0b\x32\x32.opencensus.proto.resource.v1.Resource.LabelsEntryR\x06labels\x1a\x39\n\x0bLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x98\x01\n\x1fio.opencensus.proto.resource.v1B\rResourceProtoP\x01ZEgithub.com/census-instrumentation/opencensus-proto/gen-go/resource/v1\xea\x02\x1cOpenCensus.Proto.Resource.V1b\x06proto3')



_RESOURCE = DESCRIPTOR.message_types_by_name['Resource']
_RESOURCE_LABELSENTRY = _RESOURCE.nested_types_by_name['LabelsEntry']
Resource = _reflection.GeneratedProtocolMessageType('Resource', (_message.Message,), {

  'LabelsEntry' : _reflection.GeneratedProtocolMessageType('LabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCE_LABELSENTRY,
    '__module__' : 'opencensus.proto.resource.v1.resource_pb2'
    # @@protoc_insertion_point(class_scope:opencensus.proto.resource.v1.Resource.LabelsEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCE,
  '__module__' : 'opencensus.proto.resource.v1.resource_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.resource.v1.Resource)
  })
_sym_db.RegisterMessage(Resource)
_sym_db.RegisterMessage(Resource.LabelsEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037io.opencensus.proto.resource.v1B\rResourceProtoP\001ZEgithub.com/census-instrumentation/opencensus-proto/gen-go/resource/v1\352\002\034OpenCensus.Proto.Resource.V1'
  _RESOURCE_LABELSENTRY._options = None
  _RESOURCE_LABELSENTRY._serialized_options = b'8\001'
  _RESOURCE._serialized_start=78
  _RESOURCE._serialized_end=243
  _RESOURCE_LABELSENTRY._serialized_start=186
  _RESOURCE_LABELSENTRY._serialized_end=243
# @@protoc_insertion_point(module_scope)
