from ray.util.accelerators.accelerators import (
    NVIDIA_TESLA_V100,
    NVIDIA_TESLA_P100,
    NVIDIA_TESLA_T4,
    NVIDIA_TESLA_P4,
    NVIDIA_TESLA_K80,
    NVIDIA_TESLA_A100,
    NVIDIA_TESLA_A10G,
)

__all__ = [
    "NVIDIA_TESLA_V100",
    "NVIDIA_TESLA_P100",
    "NVIDIA_TESLA_T4",
    "NVIDIA_TESLA_P4",
    "NVIDIA_TESLA_K80",
    "NVIDIA_TESLA_A100",
    "NVIDIA_TESLA_A10G",
]
