# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/agent_manager.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$src/ray/protobuf/agent_manager.proto\x12\x07ray.rpc\"z\n\x14RegisterAgentRequest\x12\x19\n\x08\x61gent_id\x18\x01 \x01(\x05R\x07\x61gentId\x12\x1d\n\nagent_port\x18\x02 \x01(\x05R\tagentPort\x12(\n\x10\x61gent_ip_address\x18\x03 \x01(\tR\x0e\x61gentIpAddress\"E\n\x12RegisterAgentReply\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x17.ray.rpc.AgentRpcStatusR\x06status*F\n\x0e\x41gentRpcStatus\x12\x17\n\x13\x41GENT_RPC_STATUS_OK\x10\x00\x12\x1b\n\x17\x41GENT_RPC_STATUS_FAILED\x10\x01\x32\x62\n\x13\x41gentManagerService\x12K\n\rRegisterAgent\x12\x1d.ray.rpc.RegisterAgentRequest\x1a\x1b.ray.rpc.RegisterAgentReplyB\x03\xf8\x01\x01\x62\x06proto3')

_AGENTRPCSTATUS = DESCRIPTOR.enum_types_by_name['AgentRpcStatus']
AgentRpcStatus = enum_type_wrapper.EnumTypeWrapper(_AGENTRPCSTATUS)
AGENT_RPC_STATUS_OK = 0
AGENT_RPC_STATUS_FAILED = 1


_REGISTERAGENTREQUEST = DESCRIPTOR.message_types_by_name['RegisterAgentRequest']
_REGISTERAGENTREPLY = DESCRIPTOR.message_types_by_name['RegisterAgentReply']
RegisterAgentRequest = _reflection.GeneratedProtocolMessageType('RegisterAgentRequest', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERAGENTREQUEST,
  '__module__' : 'src.ray.protobuf.agent_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RegisterAgentRequest)
  })
_sym_db.RegisterMessage(RegisterAgentRequest)

RegisterAgentReply = _reflection.GeneratedProtocolMessageType('RegisterAgentReply', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERAGENTREPLY,
  '__module__' : 'src.ray.protobuf.agent_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RegisterAgentReply)
  })
_sym_db.RegisterMessage(RegisterAgentReply)

_AGENTMANAGERSERVICE = DESCRIPTOR.services_by_name['AgentManagerService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _AGENTRPCSTATUS._serialized_start=244
  _AGENTRPCSTATUS._serialized_end=314
  _REGISTERAGENTREQUEST._serialized_start=49
  _REGISTERAGENTREQUEST._serialized_end=171
  _REGISTERAGENTREPLY._serialized_start=173
  _REGISTERAGENTREPLY._serialized_end=242
  _AGENTMANAGERSERVICE._serialized_start=316
  _AGENTMANAGERSERVICE._serialized_end=414
# @@protoc_insertion_point(module_scope)
